#!/usr/bin/python2.6

import sys, optparse, os, time
import Debug, GPUMultiprocessor, Timing
import atexit

# The command-line parser and its options
cmdline = optparse.OptionParser(add_help_option=False)

cmdline.add_option("-d",
                  "--debug",
                  action="store",
                  dest="debug",
                  type="int",
                  help="Debug mode.",
                  default=False)

cmdline.add_option("-p",
                  "--processors",
                  action="store",
                  dest="processors",
                  type="int",
                  help="Number of processors to use.",
                  default=1)

cmdline.add_option("-C",
                  "--deep-clean",
                  action="store_true",
                  dest="deepclean",
                  help="Clean previous runs of GPGPU-sim and uDrawgraph files.",
                  default=False)

cmdline.add_option("-h",
                  "--help",
                  action="help",
                  help="Display this help message.")

cmdline.add_option("-v",
                 "--verbose",
                 action="store_true",
                 dest="verbose",
                 help="Be verbose.",
                 default=False)

cmdline.add_option("-u",
                 "--udraw",
                 action="store_true",
                 dest="udraw",
                 help="Generate uDrawGraph files.",
                 default=False)

cmdline.add_option("-T",
                  "--number-of-tests",
                  action="store",
                  type="int",
                  dest="tests",
                  help="The number of times to run the kernel. [Default is %default].",
                  default=0,
                  metavar="<INT>")

cmdline.add_option("--no-parsing",
                 action="store_true",
                 dest="noParse",
                 help="Do not parse traces.",
                 default=False)

opts, args    = cmdline.parse_args(sys.argv[1:])
Debug.verbose = opts.verbose
Debug.debug   = opts.debug
gpgpuFileExt  = '.gpgpusim'
runPrefix     = 'run'

def analyseWarpAndWaveForIPG (program, ipg, maxWarpStartUpDelay, maxNumberOfWarps, numberOfWaves, maxLastWarpStartUpDelay):
    program.maxWarpStartUpDelay[ipg] = max(maxWarpStartUpDelay, program.maxWarpStartUpDelay[ipg])
    program.maxNumberOfWarps[ipg] = max(maxNumberOfWarps, program.maxNumberOfWarps[ipg])
    program.maxNumberOfWaves[ipg] = max(numberOfWaves, program.maxNumberOfWaves[ipg])
    program.maxLastWarpStartUpDelay[ipg] = max(maxLastWarpStartUpDelay, program.maxLastWarpStartUpDelay[ipg])

def calculateConcurrencyParameters (program, multiprocessors): 
    for SMID in multiprocessors.keys():
        multiprocessor = multiprocessors[SMID]
        Debug.debugMessage("*** Analysing condensed trace of SM %d ***" % SMID, 1)
        warpIDToPool       = {}
        warpPoolToFinished = {}
        warpPools          = []
        poolID             = 1
        warpPool           = GPUMultiprocessor.WarpPool(poolID)
        warpPoolToFinished[poolID] = False
        warpPools.append(warpPool)
        maxWarpStartUpDelay    = 0
        maxNumberOfWarps       = 0
        numberOfWaves          = 0
        precedingWarpStartTime = None
        currentIPG             = None
        lastTime               = 0
        firstWarpTime          = 0
        lastWarpTime           = 0
        for t in multiprocessor.getCondensedTrace():
            (warpID, ipointID, time) = t
            if program.isEntryIpointID(ipointID):
                ipg = program.getIPGWithEntryIpointID(ipointID)
                if currentIPG != ipg:
                    if currentIPG:
                        analyseWarpAndWaveForIPG(program, currentIPG, maxWarpStartUpDelay, maxNumberOfWarps, numberOfWaves, lastWarpTime - firstWarpTime)
                        maxWarpStartUpDelay = 0
                        maxNumberOfWarps    = 0
                        numberOfWaves       = 0
                        firstWarpTime       = time
                        lastWarpTime        = time
                    else:
                        firstWarpTime       = time
                        lastWarpTime        = time
                    currentIPG = ipg
                    Debug.debugMessage("Switching to kernel %s" % currentIPG.getName(), 1)   
                else:
                    if time < lastTime:
                        analyseWarpAndWaveForIPG(program, currentIPG, maxWarpStartUpDelay, maxNumberOfWarps, numberOfWaves, lastWarpTime - firstWarpTime)
                        maxWarpStartUpDelay = 0
                        maxNumberOfWarps    = 0
                        numberOfWaves       = 0
                        firstWarpTime       = time
                        lastWarpTime        = time
                    lastWarpTime = time
                
                if not precedingWarpStartTime:
                    precedingWarpStartTime = time
                else:
                    maxWarpStartUpDelay    = max(maxWarpStartUpDelay, time - precedingWarpStartTime)
                    precedingWarpStartTime = time
                warpPool.add(warpID)
                warpIDToPool[warpID] = poolID
                maxNumberOfWarps = max(maxNumberOfWarps, len(warpPool.warps))
            else:
                precedingWarpStartTime = None
                warpPoolID = warpIDToPool[warpID]
                del warpIDToPool[warpID]
                if not warpPoolToFinished[warpPoolID]:
                    warpPoolToFinished[poolID] = True
                    numberOfWaves += 1
                    poolID += 1
                    warpPool = GPUMultiprocessor.WarpPool(poolID)
                    warpPoolToFinished[poolID] = False
                    warpPools.append(warpPool)
            lastTime = time
        analyseWarpAndWaveForIPG(program, currentIPG, maxWarpStartUpDelay, maxNumberOfWarps, numberOfWaves, lastWarpTime - firstWarpTime)

def createGraphs (program, basepath):
    import UDrawGraph, ICFGs, Trees, IPGs
    Debug.debugMessage("Creating data structures", 1)
    for cfg in program.getCFGs():
        functionName = cfg.getName()
        UDrawGraph.makeUdrawFile (cfg, basepath, "%s.%s" % (functionName, "cfg"))
        predomTree  = Trees.Dominators(cfg, cfg.getEntryID())
        reverseg    = cfg.getReverseCFG()
        postdomTree = Trees.Dominators(reverseg, reverseg.getEntryID())
        UDrawGraph.makeUdrawFile (predomTree, basepath, "%s.%s" % (functionName, "pre"))
        UDrawGraph.makeUdrawFile (postdomTree, basepath, "%s.%s" % (functionName, "post"))
        icfg = ICFGs.ICFG(cfg)
        icfg.setEntryID()
        icfg.setExitID()
        icfg.addExitEntryEdge()
        program.addICFG(icfg)
        UDrawGraph.makeUdrawFile (icfg, basepath, "%s.%s" % (functionName, "icfg"))
        lnt = Trees.LoopNests(icfg, icfg.getEntryID())
        program.addLNT(lnt)
        UDrawGraph.makeUdrawFile (lnt, basepath, "%s.%s" % (functionName, "lnt"))
        ipg = IPGs.IPG(icfg, lnt)
        program.addIPG(ipg)
        icfg.addBranchDivergenceEdges(lnt)
        ipg.updateWithBranchDivergentPaths()
        UDrawGraph.makeUdrawFile (icfg, basepath, "%s.%s" % (functionName, "icfg"))
        UDrawGraph.makeUdrawFile (ipg, basepath, "%s.%s" % (functionName, "ipg"))

def getLineOfTimingTrace (line):
    import shlex
    tokenizer = shlex.shlex(line, posix=True)
    tokenizer.whitespace += ','
    lexemes = list(tokenizer)
    PCIndex    = 0 
    warpIndex  = 1
    SMIndex    = 2
    cycleIndex = 3
    return int(lexemes[SMIndex]), int(lexemes[warpIndex]), int(lexemes[PCIndex], 0), long(lexemes[cycleIndex])

def analyseHWMT (program, firstTuple, lastTuple):
    Debug.debugMessage("Analysing start tuple %s and last tuple %s" % (firstTuple, lastTuple), 1)
    ipointID     = firstTuple[0]
    startTime    = firstTuple[1]
    endTime      = lastTuple[1]
    ipg          = program.getIPGWithEntryIpointID(ipointID)
    functionName = ipg.getName()
    end2end      = endTime - startTime
    Debug.debugMessage("End-to-end(%s) = %d" % (functionName, end2end), 1)
    program.updateHWMT(functionName, end2end)

def splitTraces (program, generatedFiles):
    import GPUMultiprocessor
    Debug.debugMessage("Splitting traces", 1)
    multiprocessors = {}
    for outfile in generatedFiles:
        traceFound = False
        firstTuple = None
        lastTuple  = None
        Debug.debugMessage("Analysing file '%s'" % outfile, 1)
        with open(outfile, 'r') as f:
            for line in f:
                if line.startswith("NEW KERNEL"):
                        traceFound = True
                        if firstTuple:
                            analyseHWMT(program, firstTuple, lastTuple)
                            firstTuple = None
                        continue
                if traceFound:
                        SMID, warpID, ipointID, time = getLineOfTimingTrace(line)
                        timingTuple = (ipointID, time)
                        if SMID not in multiprocessors:
                            multiprocessors[SMID] = GPUMultiprocessor.StreamingMultiprocessor(SMID)
                        if not multiprocessors[SMID].hasWarp(warpID):
                            multiprocessors[SMID].addWarp(warpID)
                        multiprocessors[SMID].addToWarpTrace(warpID, timingTuple)
                        if program.isEntryIpointID(ipointID) or program.isExitIpointID(ipointID):
                            multiprocessors[SMID].addToCondensedTrace(warpID, ipointID, time)
                        lastTuple = timingTuple
                        if not firstTuple:
                            firstTuple = timingTuple 
            analyseHWMT(program, firstTuple, lastTuple)
    return multiprocessors

def writeTraces (multiprocessors, basename, basepath):
    traceFileName = basepath + os.sep + basename + ".warps.trace"
    with open(traceFileName, 'w') as f:
        for SMID in multiprocessors.keys():
            for warpID, warpTrace in multiprocessors[SMID]: 
                f.write("\n%s\n" % ('=' * 20))
                f.write("SM = %d WARP = %d\n" % (SMID, warpID))
                f.write("%s\n" % ('=' * 20))
                for t in warpTrace:
                    f.write("0x%04X %d\n" % (t[0], t[1]))
                
def createCFGs (outfile):
    import ParseCFGs
    cfgLines = []
    cfgInput = False
    with open(outfile, 'r') as f:
        for line in f:
            if line.startswith('0x'):
                break
            if "*** CFG ***" in line:
                if not cfgInput:
                    cfgInput = True
                else:
                    cfgInput = False
            if cfgInput:
                cfgLines.append(line)
    program = ParseCFGs.createProgram(cfgLines)
    return program 

def doAnalysis (generatedFiles, basename, basepath): 
    import Traces, WCET
    time1 = Timing.log("ANALYSIS BEGIN")
    # Create the CFGs
    program = createCFGs(generatedFiles[0])
    # Create the IPG
    createGraphs(program, basepath)
    if not opts.noParse:
        # Split into warp-specific traces
        multiprocessors = splitTraces (program, generatedFiles)
        if Debug.debug >= 5 and opts.tests > 0:
            writeTraces(multiprocessors, basename, basepath)
        traceData = Traces.TraceData(multiprocessors, program)
        calculateConcurrencyParameters(program, multiprocessors)
    
        for ipg in program.getIPGs():
            functionName = ipg.getName()
            # Create an ILP from the IPG and the parsed data
            ilp  = WCET.LinearProgram(ipg, traceData, functionName, basepath)
            Debug.debugMessage("********** %s **********" % functionName, 1)
            Debug.debugMessage("Warp-Specific WCET = %d" % ilp.getWCET(), 1)
            Debug.debugMessage("Warp start-up delay = %d" % program.maxWarpStartUpDelay[ipg], 1)
            Debug.debugMessage("Max number of warps = %d" % program.maxNumberOfWarps[ipg], 1)
            Debug.debugMessage("Max number of waves = %d" % program.maxNumberOfWaves[ipg], 1)
            Debug.debugMessage("Max last warp start-up delay = %d" % program.maxLastWarpStartUpDelay[ipg], 1)
            wcet = program.maxWarpStartUpDelay[ipg] * (program.maxNumberOfWarps[ipg] - 1)
            wcet = wcet + ilp.getWCET()
            wcet = program.maxNumberOfWaves[ipg] * wcet
            print "HWMT(%s) = %ld" % (functionName, program.getHWMT(functionName))
            print "WCET(%s) = %ld" % (ipg.getName(), wcet)
            print "Less pessimistic WCET(%s) = %ld" % (ipg.getName(), ilp.getWCET() + program.maxLastWarpStartUpDelay[ipg])
    time2 = Timing.log("ANALYSIS END")
    print("Analysis time = %s" %Timing.secondsToStr(time2 - time1))
            
def runCUDAKernel (basepath):
        from subprocess import Popen, PIPE
        import signal
        import re
        
        time1 = Timing.log("TESTING BEGIN")
        
        cmd = args[0]
        Debug.debugMessage("CUDA application command '%s'" % cmd, 1)
        
        run = 0
        for filename in os.listdir(basepath):
            match = re.search(r'%s[0-9]+%s' % (runPrefix, gpgpuFileExt), filename)
            if match:
                index = filename.find(gpgpuFileExt)
                num   = int(filename[len(runPrefix):index])
                if num > run:
                    run = num
        run += 1
        
        # Run the program on GPGPU-sim and get generated output
        outfiles       = []
        generatedFiles = []
        for i in xrange(run, opts.tests + run):
            outfilename = basepath + os.sep + runPrefix + str(i) + gpgpuFileExt
            generatedFiles.append(outfilename)
            outfiles.append(outfilename)
        assert generatedFiles, "No output files were generated"
        processes        = []
        pidToOutFilename = {}
        pidToOutFile     = {}
        while True:
            while outfiles and len(processes) < opts.processors:
                outfilename = outfiles.pop()
                outfile = open(outfilename, 'w')
                Debug.debugMessage("Spawning new process for '%s'" % outfilename, 1)
                proc = Popen(cmd, shell=True, stdout=outfile)
                processes.append(proc)
                pidToOutFile[proc.pid]     = outfile
                pidToOutFilename[proc.pid] = outfilename
            for p in processes:
                if p.poll() is not None:
                    processes.remove(p)
                    pidToOutFile[p.pid].close()
                    if p.returncode != 0:
                        Debug.debugMessage("Process failed for output file '%s' with return code %d" % (pidToOutFilename[p.pid], p.returncode), 1)
                        outfiles.append(pidToOutFilename[p.pid])
            if not processes and not outfiles:
                break
            else:
                time.sleep(1)
                p = Popen(['ps', '-A'], stdout=PIPE)
                out, err = p.communicate()
                for line in out.splitlines():
                    if 'cuobjdump_to_pt' in line:
                        pid = int(line.split(None, 1)[0])
                        os.kill(pid, signal.SIGKILL)
                        Debug.debugMessage("Killing stray CUDA-to-PTXPlus process", 1)
        time2 = Timing.log("TESTING END")
        Timing.secondsToStr(time2 - time1)
        print("Testing time = %s" % Timing.secondsToStr(time2 - time1))
        return generatedFiles

def checkCommandLineForAction ():    
    # Check that the user has passed the correct options
    if len(args) > 0:
        if opts.tests > 0:
            cudaBinary = args[0]
            if not os.path.exists(cudaBinary):
                Debug.exitMessage("The argument '%s' does not exist" % cudaBinary)
            elif not os.path.isfile(cudaBinary):
                Debug.exitMessage("The argument '%s' is not a file" % cudaBinary)
            elif not os.access(cudaBinary, os.X_OK):
                Debug.exitMessage("The argument '%s' does not have execute permission" % cudaBinary)
            # Get the filename of the binary without the path
            basename = os.path.basename(cudaBinary)
            basepath = os.path.abspath(os.path.dirname(cudaBinary))
            generatedFiles = runCUDAKernel(basepath)
            doAnalysis(generatedFiles, basename, basepath)
        else:
            for arg in args:
                if not arg.endswith(gpgpuFileExt):
                    Debug.exitMessage("Each file must end with a '%s' suffix. You passed '%s'." % (gpgpuFileExt, arg))
            basename = os.path.splitext(os.path.basename(args[0]))[0]
            basepath = os.path.abspath(os.path.dirname(args[0]))
            doAnalysis(args, basename, basepath)
    else:
        Debug.exitMessage("""There are two ways to run this script:
    1) Either pass a CUDA binary as an argument and the number of times you want to run the kernel with the -T option; or
    2) Pass a number of files that were generated by GPGPU-sim from a previous testing run""")

def removeFile (fullPath):
    print "Removing '%s'" % fullPath
    os.remove(fullPath)
    
def postClean (abspath):
    # The following file is generated by GPGPU-sim on every execution.
    # It includes stats about each instruction execution
    gpgpuInstrStats = abspath + os.sep + 'gpgpu_inst_stats.txt'
    if os.path.exists(gpgpuInstrStats):
        os.remove(gpgpuInstrStats)
    for paths, dirs, files in os.walk(abspath):
        files.sort()
        for filename in files:
            if filename.startswith('_cuobjdump_') or filename.startswith('_ptx'):
                removeFile(os.path.join(paths, filename))
                
def preClean (abspath):
    import re
    for paths, dirs, files in os.walk(os.path.abspath(os.curdir)):
        files.sort()
        for filename in files:
            match = re.search(r'%s[0-9]+%s' % (runPrefix, gpgpuFileExt), filename)
            if match:
                removeFile(os.path.join(paths, filename))
            elif filename.endswith('.udraw') or filename.endswith('.ilp'):
                removeFile(os.path.join(paths, filename))
            elif filename == 'benchmark.warps.trace':
                removeFile(os.path.join(paths, filename))
                
if __name__ == "__main__":
    import multiprocessing
    atexit.register(postClean, os.path.abspath(os.curdir))
    opts.processors = min(opts.processors, multiprocessing.cpu_count())
    Debug.verboseMessage("You have %d CPUs on your system. I am using %d." % (multiprocessing.cpu_count(), opts.processors))
    if opts.deepclean:
        preClean(os.path.abspath(os.curdir))
    try:
        # What to do depends on which parameters were parameters on the command line
        checkCommandLineForAction ()
    finally:
        # Remove temporarily generated files
        postClean (os.path.abspath(os.curdir))